from PyQt5.QtWidgets import QLineEdit, QListWidgetItem, QWidget, QVBoxLayout, QHBoxLayout
from PyQt5.QtWidgets import QGroupBox, QLineEdit, QLabel, QMessageBox
from PyQt5.QtCore import QSize
import sqlite3
from coolname import generate_slug
from ...factory import list_algo, list_env, get_algo, get_env
from ...utils import ystring, load_config, config_list_to_dict, normalize_routine, instantiate_env
from ...db import save_routine, remove_routine
from .variable_item import variable_item
from .objective_item import objective_item
from .constraint_item import constraint_item
from .algo_cbox import BadgerAlgoBox
from .env_cbox import BadgerEnvBox
from .config_cbox import BadgerConfigBox
from ..windows.review_dialog import BadgerReviewDialog
from ..windows.var_dialog import BadgerVariableDialog
from ..windows.edit_script_dialog import BadgerEditScriptDialog


CONS_RELATION_DICT = {
    '>': 'GREATER_THAN',
    '<': 'LESS_THAN',
    '=': 'EQUAL_TO',
}


class BadgerRoutinePage(QWidget):
    def __init__(self):
        super().__init__()

        self.algos = list_algo()
        self.envs = list_env()
        self.env = None
        self.routine = None
        self.script = ''

        self.init_ui()
        self.config_logic()

    def init_ui(self):
        # Set up the layout
        vbox = QVBoxLayout(self)
        vbox.setContentsMargins(11, 11, 19, 11)

        # Meta group
        group_meta = QGroupBox('Metadata')
        hbox_meta = QHBoxLayout(group_meta)
        label_name = QLabel('Routine Name')
        self.edit_save = edit_save = QLineEdit()
        edit_save.setPlaceholderText(generate_slug(2))
        hbox_meta.addWidget(label_name)
        hbox_meta.addWidget(edit_save, 1)
        # hbox_meta.addStretch(2)
        vbox.addWidget(group_meta)

        # Algo box
        self.algo_box = BadgerAlgoBox(self.algos)
        vbox.addWidget(self.algo_box)

        # Env box
        self.env_box = BadgerEnvBox(self.envs)
        vbox.addWidget(self.env_box)

        # Configs box
        self.configs_box = BadgerConfigBox()
        vbox.addWidget(self.configs_box)

        vbox.addStretch()

    def config_logic(self):
        self.algo_box.cb.currentIndexChanged.connect(self.select_algo)
        self.algo_box.check_use_script.stateChanged.connect(self.toggle_use_script)
        self.algo_box.btn_edit_script.clicked.connect(self.edit_script)
        self.env_box.cb.currentIndexChanged.connect(self.select_env)
        self.env_box.btn_env_play.clicked.connect(self.open_playground)
        self.configs_box.btn_all_var.clicked.connect(self.check_all_var)
        self.configs_box.btn_un_all_var.clicked.connect(self.uncheck_all_var)
        self.configs_box.btn_add_var.clicked.connect(self.add_var)
        self.configs_box.btn_all_obj.clicked.connect(self.check_all_obj)
        self.configs_box.btn_un_all_obj.clicked.connect(self.uncheck_all_obj)
        self.configs_box.check_only_var.stateChanged.connect(self.toggle_check_only_var)
        self.configs_box.check_only_obj.stateChanged.connect(self.toggle_check_only_obj)
        self.configs_box.btn_add_con.clicked.connect(self.add_constraint)

    def refresh_ui(self, routine):
        self.routine = routine  # save routine for future reference

        self.algos = list_algo()
        self.envs = list_env()
        # Clean up the constraints list
        self.configs_box.list_con.clear()

        if routine is None:
            # Reset the algo and env configs
            self.algo_box.cb.setCurrentIndex(-1)
            self.env_box.cb.setCurrentIndex(-1)

            # Reset the routine configs check box status
            self.configs_box.check_only_var.setChecked(False)
            self.configs_box.check_only_obj.setChecked(False)

            # Reset the save settings
            name = generate_slug(2)
            self.edit_save.setText('')
            self.edit_save.setPlaceholderText(name)

            return

        # Fill in the algo and env configs
        name_algo = routine['algo']
        idx_algo = self.algos.index(name_algo)
        self.algo_box.cb.setCurrentIndex(idx_algo)
        self.algo_box.edit.setPlainText(ystring(routine['algo_params']))
        try:
            self.script = routine['config']['script']
        except KeyError:
            self.script = None

        name_env = routine['env']
        idx_env = self.envs.index(name_env)
        self.env_box.cb.setCurrentIndex(idx_env)
        self.env_box.edit.setPlainText(ystring(routine['env_params']))

        # Config the vocs panel
        self.configs_box.check_only_var.setChecked(True)
        variables = [next(iter(v)) for v in routine['config']['variables']]
        for i in range(self.configs_box.list_var.count()):
            item = self.configs_box.list_var.item(i)
            item_widget = self.configs_box.list_var.itemWidget(item)
            var_name = item_widget.check_name.text()
            try:
                idx = variables.index(var_name)
                vrange = routine['config']['variables'][idx][var_name]
                item_widget.sb_lower.sb.setValue(vrange[0])
                item_widget.sb_upper.sb.setValue(vrange[1])
                item_widget.check_name.setChecked(True)
            except ValueError:
                item_widget.check_name.setChecked(False)

        self.configs_box.check_only_obj.setChecked(True)
        objectives = [next(iter(v)) for v in routine['config']['objectives']]
        for i in range(self.configs_box.list_obj.count()):
            item = self.configs_box.list_obj.item(i)
            item_widget = self.configs_box.list_obj.itemWidget(item)
            obj_name = item_widget.check_name.text()
            try:
                idx = objectives.index(obj_name)
                rule = routine['config']['objectives'][idx][obj_name]
                idx_rule = 0 if rule == 'MINIMIZE' else 1
                item_widget.cb_rule.setCurrentIndex(idx_rule)
                item_widget.check_name.setChecked(True)
            except ValueError:
                item_widget.check_name.setChecked(False)

        if routine['config']['constraints'] is not None:
            for i in range(len(routine['config']['constraints'])):
                name = next(iter(routine['config']['constraints'][i]))
                relation, thres = routine['config']['constraints'][i][name][:2]
                try:
                    routine['config']['constraints'][i][name][2]
                    critical = True
                except:
                    critical = False
                relation = ['GREATER_THAN', 'LESS_THAN',
                            'EQUAL_TO'].index(relation)
                self.add_constraint(name, relation, thres, critical)

        # Config the save settings
        name = routine['name']
        self.edit_save.setPlaceholderText(generate_slug(2))
        self.edit_save.setText(name)

        self.algo_box.check_use_script.setChecked(not not self.script)

    def select_algo(self, i):
        # Reset the script
        self.script = ''
        self.algo_box.check_use_script.setChecked(False)

        if i == -1:
            self.algo_box.edit.setPlainText('')
            return

        name = self.algos[i]
        try:
            _, configs = get_algo(name)
            self.algo_box.edit.setPlainText(ystring(configs['params']))
        except Exception as e:
            self.algo_box.cb.setCurrentIndex(-1)
            return QMessageBox.critical(self, 'Error!', str(e))

    def toggle_use_script(self):
        if self.algo_box.check_use_script.isChecked():
            self.algo_box.btn_edit_script.show()
            self.algo_box.edit.setReadOnly(True)
            self.refresh_params_algo()
        else:
            self.algo_box.btn_edit_script.hide()
            self.algo_box.edit.setReadOnly(False)

    def edit_script(self):
        algo = self.algo_box.cb.currentText()
        dlg = BadgerEditScriptDialog(self, algo, self.script, self.script_updated)
        dlg.exec()

    def script_updated(self, text):
        self.script = text
        self.refresh_params_algo()

    def refresh_params_algo(self):
        if not self.script:
            return

        try:
            tmp = {}
            exec(self.script, tmp)
            try:
                tmp['generate']  # test if generate function is defined
            except Exception as e:
                QMessageBox.warning(self, 'Please define a valid generate function!', str(e))
                return

            env_params = load_config(self.env_box.edit.toPlainText())
            try:
                intf_name = self.configs['interface'][0]
            except KeyError:
                intf_name = None
            configs = {
                'params': env_params,
                'interface': [intf_name]
            }
            env = instantiate_env(self.env, configs)
            # Function generate comes from the script
            params_algo = tmp['generate'](env, None)
            self.algo_box.edit.setPlainText(ystring(params_algo))
        except Exception as e:
            QMessageBox.warning(self, 'Invalid script!', str(e))

    def select_env(self, i):
        if i == -1:
            self.env_box.edit.setPlainText('')
            self.configs_box.list_var.clear()
            self.configs_box.dict_var.clear()
            self.configs_box.list_obj.clear()
            self.configs_box.dict_obj.clear()
            self.configs = None
            self.env = None
            self.configs_box.btn_add_con.setDisabled(True)
            self.configs_box.btn_add_var.setDisabled(True)
            self.routine = None
            return

        name = self.envs[i]
        try:
            env, configs = get_env(name)
            self.configs = configs
            self.env = env
            self.configs_box.btn_add_con.setDisabled(False)
            self.configs_box.btn_add_var.setDisabled(False)
            if self.algo_box.check_use_script.isChecked():
                self.refresh_params_algo()
        except Exception as e:
            self.configs = None
            self.env = None
            self.env_box.cb.setCurrentIndex(-1)
            self.configs_box.btn_add_con.setDisabled(True)
            self.configs_box.btn_add_var.setDisabled(True)
            self.routine = None
            return QMessageBox.critical(self, 'Error!', str(e))

        self.env_box.edit.setPlainText(ystring(configs['params']))

        self.configs_box.list_var.clear()
        vars_env = configs['variables']
        vars_combine = [*vars_env]
        if self.routine:  # check for the temp variables in vocs
            vars_vocs = self.routine['config']['variables']
            var_names_vocs = [next(iter(var)) for var in vars_vocs]
            var_names_env = [next(iter(var)) for var in vars_env]
            env_instance = instantiate_env(env, configs)
            for name in var_names_vocs:
                if name in var_names_env:
                    continue

                _var = {}
                _var[name] = env_instance._get_vrange(name)
                vars_combine.append(_var)
        for var in vars_combine:
            name = next(iter(var))
            vrange = var[name]
            item = QListWidgetItem(self.configs_box.list_var)
            var_item = variable_item(name, vrange, self.toggle_var)
            item.setSizeHint(var_item.sizeHint())
            self.configs_box.list_var.addItem(item)
            self.configs_box.list_var.setItemWidget(item, var_item)
            self.configs_box.dict_var[name] = item

        self.configs_box.list_obj.clear()
        for obj in configs['observations']:
            item = QListWidgetItem(self.configs_box.list_obj)
            obj_item = objective_item(obj, 0, self.toggle_obj)
            item.setSizeHint(obj_item.sizeHint())
            self.configs_box.list_obj.addItem(item)
            self.configs_box.list_obj.setItemWidget(item, obj_item)
            self.configs_box.dict_obj[obj] = item

        self.configs_box.list_con.clear()
        self.configs_box.fit_content()
        self.routine = None

    def open_playground(self):
        pass

    def check_all_var(self):
        for i in range(self.configs_box.list_var.count()):
            item = self.configs_box.list_var.item(i)
            item_widget = self.configs_box.list_var.itemWidget(item)
            item_widget.check_name.setChecked(True)

    def uncheck_all_var(self):
        for i in range(self.configs_box.list_var.count()):
            item = self.configs_box.list_var.item(i)
            item_widget = self.configs_box.list_var.itemWidget(item)
            item_widget.check_name.setChecked(False)

    def add_var(self):
        # TODO: Use a cached env
        env_params = load_config(self.env_box.edit.toPlainText())
        try:
            intf_name = self.configs['interface'][0]
        except KeyError:
            intf_name = None
        configs = {
            'params': env_params,
            'interface': [intf_name]
        }

        dlg = BadgerVariableDialog(self, self.env, configs, self.add_var_to_list)
        dlg.exec()

    def add_var_to_list(self, name, min, max):
        # Check if already in the list
        ok = False
        try:
            self.configs_box.dict_var[name]
        except:
            ok = True

        if not ok:
            QMessageBox.warning(self, 'Variable already exists!', f'Variable {name} already exists!')
            return 1

        item = QListWidgetItem(self.configs_box.list_var)
        var_item = variable_item(name, [min, max], self.toggle_var)
        var_item.check_name.setChecked(True)
        item.setSizeHint(var_item.sizeHint())
        self.configs_box.list_var.addItem(item)
        self.configs_box.list_var.setItemWidget(item, var_item)
        self.configs_box.dict_var[name] = item
        self.configs_box.fit_content()

        return 0

    def check_all_obj(self):
        for i in range(self.configs_box.list_obj.count()):
            item = self.configs_box.list_obj.item(i)
            item_widget = self.configs_box.list_obj.itemWidget(item)
            item_widget.check_name.setChecked(True)

    def uncheck_all_obj(self):
        for i in range(self.configs_box.list_obj.count()):
            item = self.configs_box.list_obj.item(i)
            item_widget = self.configs_box.list_obj.itemWidget(item)
            item_widget.check_name.setChecked(False)

    def toggle_check_only_var(self):
        if self.configs_box.check_only_var.isChecked():
            for i in range(self.configs_box.list_var.count()):
                item = self.configs_box.list_var.item(i)
                item_widget = self.configs_box.list_var.itemWidget(item)
                if not item_widget.check_name.isChecked():
                    item_widget.hide()
                    item.setSizeHint(QSize(0, 0))
        else:
            for i in range(self.configs_box.list_var.count()):
                item = self.configs_box.list_var.item(i)
                item_widget = self.configs_box.list_var.itemWidget(item)
                item_widget.show()
                item.setSizeHint(item_widget.sizeHint())
        self.configs_box.fit_content()

    def toggle_var(self, name):
        if self.configs_box.check_only_var.isChecked():
            item = self.configs_box.dict_var[name]
            item_widget = self.configs_box.list_var.itemWidget(item)
            if item_widget.check_name.isChecked():
                item_widget.show()
                item.setSizeHint(item_widget.sizeHint())
            else:
                item_widget.hide()
                item.setSizeHint(QSize(0, 0))
            self.configs_box.fit_content()

    def toggle_check_only_obj(self):
        if self.configs_box.check_only_obj.isChecked():
            for i in range(self.configs_box.list_obj.count()):
                item = self.configs_box.list_obj.item(i)
                item_widget = self.configs_box.list_obj.itemWidget(item)
                if not item_widget.check_name.isChecked():
                    item_widget.hide()
                    item.setSizeHint(QSize(0, 0))
        else:
            for i in range(self.configs_box.list_obj.count()):
                item = self.configs_box.list_obj.item(i)
                item_widget = self.configs_box.list_obj.itemWidget(item)
                item_widget.show()
                item.setSizeHint(item_widget.sizeHint())
        self.configs_box.fit_content()

    def toggle_obj(self, name):
        if self.configs_box.check_only_obj.isChecked():
            item = self.configs_box.dict_obj[name]
            item_widget = self.configs_box.list_obj.itemWidget(item)
            if item_widget.check_name.isChecked():
                item_widget.show()
                item.setSizeHint(item_widget.sizeHint())
            else:
                item_widget.hide()
                item.setSizeHint(QSize(0, 0))
            self.configs_box.fit_content()

    def add_constraint(self, name=None, relation=0, threshold=0, critical=False):
        if self.configs is None:
            return

        options = self.configs['observations']
        item = QListWidgetItem(self.configs_box.list_con)
        con_item = constraint_item(options,
                                   lambda: self.configs_box.list_con.takeItem(
                                       self.configs_box.list_con.row(item)),
                                   name, relation, threshold, critical)
        item.setSizeHint(con_item.sizeHint())
        self.configs_box.list_con.addItem(item)
        self.configs_box.list_con.setItemWidget(item, con_item)
        # self.configs_box.dict_con[''] = item
        self.configs_box.fit_content()

    def _compose_routine(self):
        # Compose the routine
        name = self.edit_save.text() or self.edit_save.placeholderText()
        algo = self.algo_box.cb.currentText()
        assert algo, 'Please specify algorithm'
        env = self.env_box.cb.currentText()
        assert env, 'Please specify environment'
        algo_params = load_config(self.algo_box.edit.toPlainText())
        env_params = load_config(self.env_box.edit.toPlainText())

        variables = []
        for i in range(self.configs_box.list_var.count()):
            item = self.configs_box.list_var.item(i)
            item_widget = self.configs_box.list_var.itemWidget(item)
            if item_widget.check_name.isChecked():
                var_name = item_widget.check_name.text()
                lb = item_widget.sb_lower.sb.value()
                ub = item_widget.sb_upper.sb.value()
                _dict = {}
                _dict[var_name] = [lb, ub]
                variables.append(_dict)

        objectives = []
        for i in range(self.configs_box.list_obj.count()):
            item = self.configs_box.list_obj.item(i)
            item_widget = self.configs_box.list_obj.itemWidget(item)
            if item_widget.check_name.isChecked():
                obj_name = item_widget.check_name.text()
                rule = item_widget.cb_rule.currentText()
                _dict = {}
                _dict[obj_name] = rule
                objectives.append(_dict)

        constraints = []
        for i in range(self.configs_box.list_con.count()):
            item = self.configs_box.list_con.item(i)
            item_widget = self.configs_box.list_con.itemWidget(item)
            critical = item_widget.check_crit.isChecked()
            con_name = item_widget.cb_obs.currentText()
            relation = CONS_RELATION_DICT[item_widget.cb_rel.currentText()]
            value = item_widget.sb.value()
            _dict = {}
            _dict[con_name] = [relation, value]
            if critical:
                _dict[con_name].append('CRITICAL')
            constraints.append(_dict)

        configs = {
            'variables': variables,
            'objectives': objectives,
            'constraints': constraints,
        }
        if self.algo_box.check_use_script.isChecked():
            configs['script'] = self.script

        routine = {
            'name': name,
            'algo': algo,
            'env': env,
            'algo_params': algo_params,
            'env_params': env_params,
            'env_vranges': config_list_to_dict(variables),
            'config': configs,
        }

        # Sanity check and config normalization
        try:
            routine = normalize_routine(routine)
        except Exception as e:
            raise e

        return routine

    def review(self):
        try:
            routine = self._compose_routine()
        except Exception as e:
            return QMessageBox.critical(self, 'Error!', str(e))

        dlg = BadgerReviewDialog(self, routine)
        dlg.exec()

    def save(self):
        try:
            routine = self._compose_routine()
        except Exception as e:
            return QMessageBox.critical(self, 'Error!', str(e))

        try:
            save_routine(routine)
        except sqlite3.IntegrityError:
            return QMessageBox.critical(self, 'Error!',
                f'Routine {routine["name"]} already existed in the database! Please choose another name.')

        return 0

    def delete(self):
        name = self.edit_save.text() or self.edit_save.placeholderText()
        remove_routine(name)

        return 0



import os


def test_algo_api():
    from badger.factory import list_algo, get_algo

    assert len(list_algo()) == 2

    _, configs = get_algo('silly')
    assert configs['name'] == 'silly'
    assert configs['version'] == '0.1'

    _, configs = get_algo('simplex')
    assert configs['name'] == 'simplex'
    assert configs['version'] == '0.1'


def test_env_api():
    from badger.factory import list_env, get_env

    assert len(list_env()) == 1

    _, configs = get_env('silly')
    assert configs['name'] == 'silly'
    assert configs['version'] == '0.1'


def test_intf_api():
    from badger.factory import list_intf, get_intf

    assert len(list_intf()) == 1

    _, configs = get_intf('silly')
    assert configs['name'] == 'silly'
    assert configs['version'] == '0.1'


def test_run(mock_config_root):
    from coolname import generate_slug
    from badger.log import config_log
    config_log()
    from badger.factory import get_algo, get_env
    from badger.utils import load_config, merge_params, config_list_to_dict
    from badger.utils import normalize_routine, run_routine

    # Args
    algo = 'silly'
    algo_params = '{dimension: 2, max_iter: 10}'
    env = 'silly'
    env_params = None
    config = os.path.join(mock_config_root, 'test.yaml')
    save = False
    yes = True
    verbose = 2

    # Get env params
    _, configs_env = get_env(env)

    # Get algo params
    _, configs_algo = get_algo(algo)

    # Normalize the algo and env params
    params_env = load_config(env_params)
    params_algo = load_config(algo_params)
    params_env = merge_params(configs_env['params'], params_env)
    params_algo = merge_params(configs_algo['params'], params_algo)

    # Load routine configs
    configs_routine = load_config(config)

    # Compose the routine
    routine = {
        'name': save or generate_slug(2),
        'algo': algo,
        'env': env,
        'algo_params': params_algo,
        'env_params': params_env,
        # env_vranges is an additional info for the normalization
        # Will be removed after the normalization
        'env_vranges': config_list_to_dict(configs_env['variables']),
        'config': configs_routine,
    }

    # Sanity check and config normalization
    routine = normalize_routine(routine)

    run_routine(routine, yes, save, verbose)



from PyQt5.QtWidgets import QComboBox, QGridLayout, QVBoxLayout, QWidget, QLabel
from PyQt5.QtWidgets import QDialog, QDialogButtonBox, QApplication, QStyledItemDelegate
from PyQt5.QtCore import QSettings
from qdarkstyle import load_stylesheet, DarkPalette, LightPalette


class BadgerSettingsDialog(QDialog):
    theme_list = ['default', 'light', 'dark']
    theme_idx_dict = {
        'default': 0,
        'light': 1,
        'dark': 2,
    }

    def __init__(self, parent):
        super().__init__(parent)

        self.settings = QSettings('SLAC-ML', 'Badger')
        self._settings = {}  # previous settings
        keys = self.settings.allKeys()
        for key in keys:
            self._settings[key] = self.settings.value(key)

        self.init_ui()
        self.config_logic()

    def init_ui(self):
        self.setWindowTitle('Badger settings')
        self.resize(320, 240)

        vbox = QVBoxLayout(self)

        widget_settings = QWidget(self)
        grid = QGridLayout(widget_settings)
        grid.setContentsMargins(0, 0, 0, 0)

        # Theme selector
        theme = self.settings.value('theme')
        self.lbl_theme = lbl_theme = QLabel('Theme')
        self.cb_theme = cb_theme = QComboBox()
        cb_theme.setItemDelegate(QStyledItemDelegate())
        cb_theme.addItems(self.theme_list)
        cb_theme.setCurrentIndex(self.theme_idx_dict[theme])

        grid.addWidget(lbl_theme, 0, 0)
        grid.addWidget(cb_theme, 0, 1)

        grid.setColumnStretch(1, 1)

        vbox.addWidget(widget_settings)

        self.btns = QDialogButtonBox(
            QDialogButtonBox.Ok | QDialogButtonBox.Cancel)

        vbox.addStretch(1)
        vbox.addWidget(self.btns)

    def config_logic(self):
        self.cb_theme.currentIndexChanged.connect(self.select_theme)
        self.btns.accepted.connect(self.apply_settings)
        self.btns.rejected.connect(self.restore_settings)

    def select_theme(self, i):
        app = QApplication.instance()
        theme = self.theme_list[i]
        if theme == 'dark':
            app.setStyleSheet(load_stylesheet(palette=DarkPalette))
        elif theme == 'light':
            app.setStyleSheet(load_stylesheet(palette=LightPalette))
        else:
            app.setStyleSheet('')
        self.settings.setValue('theme', theme)

    def apply_settings(self):
        self.accept()

    def restore_settings(self):
        app = QApplication.instance()
        for key in self._settings.keys():
            self.settings.setValue(key, self._settings[key])

        theme = self.settings.value('theme')
        if theme == 'dark':
            app.setStyleSheet(load_stylesheet(palette=DarkPalette))
        elif theme == 'light':
            app.setStyleSheet(load_stylesheet(palette=LightPalette))
        else:
            app.setStyleSheet('')

        self.reject()



from pkg_resources import resource_filename
import signal
import time
from PyQt5.QtWidgets import QApplication
from PyQt5.QtGui import QFont, QIcon
from PyQt5 import QtCore
import sys
# import ctypes
from qdarkstyle import load_stylesheet, LightPalette, DarkPalette
from .windows.main_window import BadgerMainWindow

# Fix the scaling issue on multiple monitors w/ different scaling settings
# if sys.platform == 'win32':
#     ctypes.windll.shcore.SetProcessDpiAwareness(1)

if hasattr(QtCore.Qt, 'AA_EnableHighDpiScaling'):
    QApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling, True)

if hasattr(QtCore.Qt, 'AA_UseHighDpiPixmaps'):
    QApplication.setAttribute(QtCore.Qt.AA_UseHighDpiPixmaps, True)

# if hasattr(QtCore.Qt, 'HighDpiScaleFactorRoundingPolicy'):
#     QApplication.setAttribute(
#         QtCore.Qt.HighDpiScaleFactorRoundingPolicy.PassThrough)

TIMER = {'time': None}


def on_exit(*args):
    if TIMER['time'] is None:
        print('Press Ctrl/Cmd + C again within 1s to quit Badger')
        TIMER['time'] = time.time()
        return

    TIMER['time'] = None
    print('Goodbye')
    QApplication.quit()


def on_timeout():
    if TIMER['time'] is None:
        return

    now = time.time()
    if (now - TIMER['time']) > 1:
        TIMER['time'] = None
        print('Timeout, resume the operation...')


def launch_gui():
    app = QApplication(sys.argv)

    # Set app metainfo
    app.setApplicationName('Badger')
    icon_path = resource_filename(__name__, 'images/icon.png')
    app.setWindowIcon(QIcon(icon_path))

    font = QFont()
    font.setPixelSize(13)
    # font.setWeight(QFont.DemiBold)
    app.setFont(font)

    # Configure app settings
    settings = QtCore.QSettings('SLAC-ML', 'Badger')
    theme = settings.value('theme')
    if not theme:
        settings.setValue('theme', 'dark')
        theme = settings.value('theme')

    # Set up stylesheet
    if theme == 'dark':
        app.setStyleSheet(load_stylesheet(palette=DarkPalette))
    elif theme == 'light':
        app.setStyleSheet(load_stylesheet(palette=LightPalette))
    else:
        app.setStyleSheet('')

    # Show the main window
    window = BadgerMainWindow()

    # Enable Ctrl + C quit
    signal.signal(signal.SIGINT, on_exit)
    # Let the interpreter run each 0.2 s
    timer = QtCore.QTimer()
    timer.timeout.connect(on_timeout)
    timer.start(200)

    window.show()
    sys.exit(app.exec())



from setuptools import setup, find_packages
import versioneer

setup(
    name='badger-opt',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description='Core of the Badger optimizer',
    url='https://github.com/SLAC-ML/Badger',
    author='Zhe Zhang',
    author_email='zhezhang@slac.stanford.edu',
    license='GPL',
    packages=find_packages(where='src'),
    package_dir={'': 'src'},
    include_package_data=True,
    install_requires=[
        'pandas',
        'pyyaml',
        'coolname',
        'pyqt5',
        'pyqtgraph',
        'qdarkstyle',
        'pillow'
    ],
    extras_require={
        'dev': [
            'pytest',
            'pytest-qt'
        ]
    },
    python_requires='>=3.6',
    entry_points={
        'console_scripts': [
            'badger = badger.__main__:main'
        ]
    },
)



import os
from datetime import datetime
import logging
logger = logging.getLogger(__name__)
from .settings import read_value


# Check badger logbook root
BADGER_LOGBOOK_ROOT = read_value('BADGER_LOGBOOK_ROOT')
if BADGER_LOGBOOK_ROOT is None:
    raise Exception('Please set the BADGER_LOGBOOK_ROOT env var!')
elif not os.path.exists(BADGER_LOGBOOK_ROOT):
    os.makedirs(BADGER_LOGBOOK_ROOT)
    logger.info(
        f'Badger logbook root {BADGER_LOGBOOK_ROOT} created')


def send_to_logbook(routine, data, widget=None):
    from xml.etree import ElementTree
    from re import sub

    log_text = ''
    algo_name = 'simplex'
    data_path = 'runs'
    obj_name = 'obj'
    obj_start = 0
    obj_end = 1
    n_iter = 10
    n_point = 10
    if n_iter > 0:
        log_text = f'Gain ({obj_name}): {round(obj_start, 4)} -> {round(obj_end, 4)}'
    log_text += f'\nIterations: {n_iter}\n'
    # log_text += 'Trim delay: '+str(gui.sb_tdelay.value())+'\n'
    log_text += f'Points Requested: {n_point}\n'
    # log_text += 'Normalization Amp Coeff: ' + \
        # str(gui.sb_scaling_coef.value())+'\n'
    log_text += f'Type of optimization: {algo_name}\n'
    log_text += f'Data location: {data_path}\n'
    try:
        log_text += f'Log location: {BADGER_LOGBOOK_ROOT}\n'
    except:
        pass

    # Generate the xml data
    curr_time = datetime.now()
    if os.name == 'nt':
        timestr = curr_time.strftime('%Y-%m-%dT%H%M%S')
    else:
        timestr = curr_time.strftime('%Y-%m-%dT%H:%M:%S')
    log_entry = ElementTree.Element(None)
    severity = ElementTree.SubElement(log_entry, 'severity')
    location = ElementTree.SubElement(log_entry, 'location')
    keywords = ElementTree.SubElement(log_entry, 'keywords')
    time = ElementTree.SubElement(log_entry, 'time')
    isodate = ElementTree.SubElement(log_entry, 'isodate')
    log_user = ElementTree.SubElement(log_entry, 'author')
    category = ElementTree.SubElement(log_entry, 'category')
    title = ElementTree.SubElement(log_entry, 'title')
    metainfo = ElementTree.SubElement(log_entry, 'metainfo')
    imageFile = ElementTree.SubElement(log_entry, 'link')
    imageFile.text = timestr + '-00.ps'
    thumbnail = ElementTree.SubElement(log_entry, 'file')
    thumbnail.text = timestr + '-00.png'
    text = ElementTree.SubElement(log_entry, 'text')
    log_entry.attrib['type'] = 'LOGENTRY'
    category.text = 'USERLOG'
    location.text = 'not set'
    severity.text = 'NONE'
    keywords.text = 'none'
    time.text = curr_time.strftime('%H:%M:%S')
    isodate.text = curr_time.strftime('%Y-%m-%d')
    metainfo.text = timestr + '-00.xml'
    log_user.text = ' '
    title.text = 'Badger'
    text.text = log_text
    if text.text == '':
        text.text = ' '  # If field is truly empty, ElementTree leaves off tag entirely which causes logbook parser to fail

    fileName = os.path.join(BADGER_LOGBOOK_ROOT, metainfo.text)
    fileName = fileName.rstrip('.xml')
    xmlFile = open(fileName + '.xml', 'w')
    rawString = ElementTree.tostring(log_entry, 'utf-8').decode('utf-8')
    parsedString = sub(r'(?=<[^/].*>)', '\n', rawString)
    xmlString = parsedString[1:]
    xmlFile.write(xmlString)
    xmlFile.write('\n')  # Close with newline so cron job parses correctly
    xmlFile.close()
    screenshot(widget, f'{fileName}.png')


def screenshot(widget, filename):
    """
    Takes a screenshot of the whole gui window, saves png and ps images to file
    """
    if widget is None:
        raise Exception('No widget to take screenshot on!')

    from PIL import Image

    pic = widget.grab()
    pic.save(filename)
    img = Image.open(filename)
    if img.mode in ('RGBA', 'LA'):
        # https://pillow.readthedocs.io/en/stable/handbook/image-file-formats.html?highlight=eps#eps
        # logger.warning(f'Current figure mode "{img.mode}" cannot be directly saved to .ps and will be converted to "RGB" mode')
        img = img.convert('RGB')
    # img = img.scaled(400, 600)
    name = os.path.splitext(filename)[0]
    img.save(f'{name}.ps')



from ..settings import read_value, BADGER_PATH_DICT
from .config import _config_path_var


def self_check(args):
    good = check_n_config_paths()
    if good:
        print('Badger is healthy!')


def check_n_config_paths():
    good = True

    for pname in BADGER_PATH_DICT.keys():
        if not read_value(pname):
            good = False
            dname = BADGER_PATH_DICT[pname]['display name']
            print(f'\n{dname} needs to be configured!\n')

            try:
                _config_path_var(pname)
            except KeyboardInterrupt:
                pass

    return good



from logging.config import dictConfig
from .utils import merge_params


LOG_CONFIG_DEFAULT = {
    'version': 1,
    'formatters': {
        'simple': {
            'format': '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
        }
    },
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
            'level': 'DEBUG',
            'formatter': 'simple',
            'stream': 'ext://sys.stdout'
        }
    },
    'root': {
        'level': 'WARNING',
        'handlers': [
            'console'
        ]
    }
}


def config_log(dict_config=None):
    _config = merge_params(LOG_CONFIG_DEFAULT, dict_config)
    dictConfig(_config)



from pkg_resources import get_distribution
try:
    from ..factory import BADGER_PLUGIN_ROOT, BADGER_EXTENSIONS
except:
    BADGER_PLUGIN_ROOT = None
    BADGER_EXTENSIONS = None
try:
    from ..db import BADGER_DB_ROOT
except:
    BADGER_DB_ROOT = None
try:
    from ..logbook import BADGER_LOGBOOK_ROOT
except:
    BADGER_LOGBOOK_ROOT = None
try:
    from ..archive import BADGER_ARCHIVE_ROOT
except:
    BADGER_ARCHIVE_ROOT = None
from ..utils import yprint
from .doctor import check_n_config_paths


def show_info(args):
    if args.gui:
        from ..gui import launch_gui

        launch_gui()
    elif args.gui_acr:
        from ..gui_acr import launch_gui

        launch_gui()
    else:
        info = {
            'name': 'Badger the optimizer',
            'version': get_distribution('badger-opt').version,
            'plugin root': BADGER_PLUGIN_ROOT,
            'database root': BADGER_DB_ROOT,
            'logbook root': BADGER_LOGBOOK_ROOT,
            'archive root': BADGER_ARCHIVE_ROOT,
        }

        if BADGER_EXTENSIONS:
            extensions = list(BADGER_EXTENSIONS.keys())
            if extensions:
                info['extensions'] = extensions

        yprint(info)

        check_n_config_paths()
        # print(f'Badger the optimizer')
        # print('====================')
        # print(f'version      : {version}')
        # print(f'plugins root : {BADGER_PLUGIN_ROOT}')



from ..settings import list_settings, read_value, write_value, BADGER_PATH_DICT
import os
import logging
logger = logging.getLogger(__name__)
from ..utils import yprint


def config_settings(args):
    key = args.key

    if key is None:
        yprint(list_settings())
        return

    try:
        print('')
        return _config_path_var(key)
    except IndexError:
        pass
    except KeyboardInterrupt:
        return

    logger.error(f'{key} is not a valid Badger config key!')


def _config_path_var(var_name):
    display_name = BADGER_PATH_DICT[var_name]['display name']
    desc = BADGER_PATH_DICT[var_name]['description']

    print(f'=== Configure {display_name} ===')
    print(f'*** {desc} ***\n')
    while True:
        res = input(
            f'Please type in the path to the Badger {display_name} folder (S to skip, R to reset): \n')
        if res == 'S':
            break
        if res == 'R':
            _res = input(
                f'The current value {read_value(var_name)} will be reset, proceed (y/[n])? ')
            if _res == 'y':
                break
            elif (not _res) or (_res == 'n'):
                print('')
                continue
            else:
                print(f'Invalid choice: {_res}')

        res = os.path.abspath(os.path.expanduser(res))
        if os.path.isdir(res):
            _res = input(f'Your choice is {res}, proceed ([y]/n)? ')
            if _res == 'n':
                print('')
                continue
            elif (not _res) or (_res == 'y'):
                break
            else:
                print(f'Invalid choice: {_res}')
        else:
            _res = input(
                f'{res} does not exist, do you want to create it ([y]/n)? ')
            if _res == 'n':
                print('')
                continue
            elif (not _res) or (_res == 'y'):
                os.makedirs(res)
                print(f'Directory {res} has been created')
                break
            else:
                print(f'Invalid choice: {_res}')

    if res == 'R':
        write_value(var_name, None)
        print(f'You reset the Badger {display_name} folder setting')
    elif res != 'S':
        write_value(var_name, res)
        print(f'You set the Badger {display_name} folder to {res}')



import os
import shutil
import pytest


@pytest.fixture(scope='module', autouse=True)
def config_test_settings(mock_plugin_root, mock_db_root,
                         mock_logbook_root, mock_archive_root):
    from badger.settings import read_value, write_value

    # Store the old values
    old_root = read_value('BADGER_PLUGIN_ROOT')
    old_db = read_value('BADGER_DB_ROOT')
    old_logbook = read_value('BADGER_LOGBOOK_ROOT')
    old_archived = read_value('BADGER_ARCHIVE_ROOT')

    # Assign values for test
    write_value('BADGER_PLUGIN_ROOT', mock_plugin_root)
    write_value('BADGER_DB_ROOT', mock_db_root)
    write_value('BADGER_LOGBOOK_ROOT', mock_logbook_root)
    write_value('BADGER_ARCHIVE_ROOT', mock_archive_root)

    yield

    # Restoring the original settings
    write_value('BADGER_PLUGIN_ROOT', old_root)
    write_value('BADGER_DB_ROOT', old_db)
    write_value('BADGER_LOGBOOK_ROOT', old_logbook)
    write_value('BADGER_ARCHIVE_ROOT', old_archived)


@pytest.fixture(scope='module', autouse=True)
def clean_up(mock_db_root, mock_logbook_root, mock_archive_root):
    # Clean before tests
    shutil.rmtree(mock_db_root, True)  # ignore errors
    shutil.rmtree(mock_logbook_root, True)
    shutil.rmtree(mock_archive_root, True)

    yield

    # Clean after tests
    shutil.rmtree(mock_db_root, True)
    shutil.rmtree(mock_logbook_root, True)
    shutil.rmtree(mock_archive_root, True)


@pytest.fixture(scope='module')
def mock_root(request):
    return os.path.join(request.fspath.dirname, 'mock')


@pytest.fixture(scope='module')
def mock_plugin_root(mock_root):
    return os.path.join(mock_root, 'plugins')


@pytest.fixture(scope='module')
def mock_db_root(mock_root):
    return os.path.join(mock_root, 'db')


@pytest.fixture(scope='module')
def mock_logbook_root(mock_root):
    return os.path.join(mock_root, 'logbook')


@pytest.fixture(scope='module')
def mock_archive_root(mock_root):
    return os.path.join(mock_root, 'archived')


@pytest.fixture(scope='module')
def mock_config_root(mock_root):
    return os.path.join(mock_root, 'configs')



import numpy as np
from operator import itemgetter
from badger.utils import ParetoFront


def optimize(evaluate, params):
    D, max_iter = itemgetter('dimension', 'max_iter')(params)

    for i in range(max_iter):
        x = np.random.rand(D).reshape(1, -1)
        y, _, _, _ = evaluate(x)
        if not i:
            pf = ParetoFront(['MINIMIZE'] * y.shape[1])

        pf.is_dominated((x, y))
    # print(len(pf.pareto_front) / max_iter)

    return pf.pareto_front, pf.pareto_set



from PyQt5.QtCore import QSettings


BADGER_PATH_DICT = {
    'BADGER_PLUGIN_ROOT': {
        'display name': 'plugin root',
        'description': 'This setting (BADGER_PLUGIN_ROOT) tells Badger where to look for the plugins',
    },
    'BADGER_DB_ROOT': {
        'display name': 'database root',
        'description': 'This setting (BADGER_DB_ROOT) tells Badger where to store the routine database',
    },
    'BADGER_LOGBOOK_ROOT': {
        'display name': 'logbook root',
        'description': 'This setting (BADGER_LOGBOOK_ROOT) tells Badger where to send the logs (GUI mode)',
    },
    'BADGER_ARCHIVE_ROOT': {
        'display name': 'archive root',
        'description': 'This setting (BADGER_ARCHIVE_ROOT) tells Badger where to archive the historical optimization runs',
    },
}


def list_settings():
    """List all the settings in Badger

    Returns
    -------
    dict
        A dictionary contains the settings. Keys in the dict are fields of the
        settings, the value for each key is the current value for that setting.

    """
    settings = QSettings('SLAC-ML', 'Badger')
    result = {}
    for key in BADGER_PATH_DICT.keys():
        result[key] = settings.value(key)

    return result


def read_value(key):
    settings = QSettings('SLAC-ML', 'Badger')

    return settings.value(key)


def write_value(key, value):
    settings = QSettings('SLAC-ML', 'Badger')

    settings.setValue(key, value)



from PyQt5.QtWidgets import QVBoxLayout, QHBoxLayout, QPushButton, QWidget, QPlainTextEdit
from PyQt5.QtWidgets import QComboBox, QCheckBox, QStyledItemDelegate, QLabel
from .collapsible_box import CollapsibleBox


class BadgerAlgoBox(CollapsibleBox):
    def __init__(self, algos=[], parent=None):
        super().__init__(' Algorithm', parent)

        self.algos = algos

        self.init_ui()

    def init_ui(self):
        vbox = QVBoxLayout()

        name = QWidget()
        hbox_name = QHBoxLayout(name)
        hbox_name.setContentsMargins(0, 0, 0, 0)
        lbl = QLabel('Name')
        lbl.setFixedWidth(64)
        self.cb = cb = QComboBox()
        cb.setItemDelegate(QStyledItemDelegate())
        cb.addItems(self.algos)
        cb.setCurrentIndex(-1)
        hbox_name.addWidget(lbl)
        hbox_name.addWidget(cb, 1)
        vbox.addWidget(name)

        params = QWidget()
        hbox_params = QHBoxLayout(params)
        hbox_params.setContentsMargins(0, 0, 0, 0)
        lbl_params_col = QWidget()
        vbox_lbl_params = QVBoxLayout(lbl_params_col)
        vbox_lbl_params.setContentsMargins(0, 0, 0, 0)
        lbl_params = QLabel('Params')
        lbl_params.setFixedWidth(64)
        vbox_lbl_params.addWidget(lbl_params)
        vbox_lbl_params.addStretch(1)
        hbox_params.addWidget(lbl_params_col)

        edit_params_col = QWidget()
        vbox_params_edit = QVBoxLayout(edit_params_col)
        vbox_params_edit.setContentsMargins(0, 0, 0, 8)
        script_bar = QWidget()
        hbox_script = QHBoxLayout(script_bar)
        hbox_script.setContentsMargins(0, 0, 0, 0)
        vbox_params_edit.addWidget(script_bar)
        self.check_use_script = check_use_script = QCheckBox('Generate from Script')
        check_use_script.setChecked(False)
        self.btn_edit_script = btn_edit_script = QPushButton('Edit Script')
        btn_edit_script.setFixedSize(96, 24)
        btn_edit_script.hide()
        hbox_script.addWidget(check_use_script)
        hbox_script.addWidget(btn_edit_script)
        self.edit = edit = QPlainTextEdit()
        edit.setFixedHeight(128)
        vbox_params_edit.addWidget(edit)
        hbox_params.addWidget(edit_params_col)
        vbox.addWidget(params)

        self.setContentLayout(vbox)
